<?php

declare(strict_types=1);

namespace Drupal\data_pipelines_sftp\Source\Resource;

use Drupal\Component\Serialization\Json;
use Drupal\Core\Field\BaseFieldDefinition;
use Drupal\Core\StringTranslation\TranslatableMarkup;
use Drupal\data_pipelines\Entity\DatasetInterface;
use Drupal\data_pipelines\Source\Resource\SourceResourceBase;
use Drupal\key\KeyRepositoryInterface;
use phpseclib3\Net\SFTP as SFTPClient;
use phpseclib3\Net\SFTP\Stream;

/**
 * Provides SFTP a source resource.
 */
final class Sftp extends SourceResourceBase {

  public function __construct(protected KeyRepositoryInterface $keyRepository) {}

  /**
   * {@inheritdoc}
   */
  public function getResource(DatasetInterface $dataset, string $field_name): mixed {
    $path = self::getFieldValue($dataset, $field_name);
    $local_copy = self::getFieldValue($dataset, $field_name, 'local_copy');
    $host = self::getFieldValue($dataset, $field_name, 'host');
    $port = (int) self::getFieldValue($dataset, $field_name, 'port');
    $credentials = self::getFieldValue($dataset, $field_name, 'credentials');
    ['username' => $username, 'password' => $password] = Json::decode($this->keyRepository->getKey($credentials)->getKeyValue());
    try {
      $client = $this->getClient($host, $port, $username, $password);
      return $local_copy ? $this->createLocalCopy($client, $path) : $this->createMemoryCopy($client, $path);
    }
    catch (\Exception $e) {
      $dataset->addLogMessage(sprintf("There was an error getting the file from the file server - %s", $e->getMessage()));
    }
    $dataset->addLogMessage(sprintf("The local copy of %s has been used.", $path));
    return $local_copy ? $this->getLocalCopy($path) : NULL;
  }

  /**
   * A method to get the authenticated SFTP client.
   *
   * @param string $host
   *   The host.
   * @param int $port
   *   The port.
   * @param string $username
   *   The username.
   * @param string $password
   *   The password.
   *
   * @return \phpseclib3\Net\SFTP
   *   The authenticated SFTP client.
   *
   * @throws \Exception
   */
  protected function getClient(string $host, int $port, string $username, string $password): SFTPClient {
    static $client;
    if ($client) {
      return $client;
    }
    Stream::register();
    $client = new SFTPClient($host, $port ?: 22);
    if (!$client->login($username, $password)) {
      throw new \Exception('Authentication failed.');
    }
    return $client;
  }

  /**
   * A method to get a local copy of the file if opted into.
   *
   * This method is used as a fallback - if the file cannot be fetched and
   * a local copy is available.
   *
   * @param string $path
   *   The file path.
   *
   * @return mixed
   *   The resource.
   */
  protected function getLocalCopy(string $path): mixed {
    $filename = pathinfo($path, PATHINFO_BASENAME);
    $local_path = 'private://' . $filename;
    if (is_readable($local_path)) {
      if ($local_stream = fopen($local_path, 'r')) {
        rewind($local_stream);
        return $local_stream;
      }
    }
    return NULL;
  }

  /**
   * A method to create a local copy of the file if opted into.
   *
   * @param \phpseclib3\Net\SFTP $client
   *   The authenticated SFTP client.
   * @param string $path
   *   The file path.
   *
   * @return string
   *   The resource.
   *
   * @throws \Exception
   */
  protected function createLocalCopy(SFTPClient $client, string $path): mixed {
    $filename = pathinfo($path, PATHINFO_BASENAME);
    $local_path = 'private://' . $filename;
    if ($local_stream = fopen($local_path, 'a+')) {
      if (!fwrite($local_stream, $client->get($path))) {
        throw new \Exception('Could not create local copy.');
      }
      rewind($local_stream);
      return $local_stream;
    }
    return NULL;
  }

  /**
   * A method to create a memory copy of the file.
   *
   * @param \phpseclib3\Net\SFTP $client
   *   The authenticated SFTP client.
   * @param string $path
   *   The file path.
   *
   * @return string
   *   The resource.
   *
   * @throws \Exception
   */
  protected function createMemoryCopy(SFTPClient $client, string $path): mixed {
    $memory_path = 'php://temp';
    if ($memory_stream = fopen($memory_path, 'w+')) {
      if (!fwrite($memory_stream, $client->get($path))) {
        throw new \Exception('Could not create memory copy.');
      }
      rewind($memory_stream);
      return $memory_stream;
    }
    return NULL;
  }

  /**
   * {@inheritdoc}
   */
  public function getResourceBaseFieldDefinition(array $source_plugin_definition): BaseFieldDefinition {
    return BaseFieldDefinition::create('sftp')
      ->setLabel(new TranslatableMarkup('SFTP details'))
      ->setRequired(TRUE)
      ->setDisplayOptions('form', [
        'type' => 'sftp',
      ]);
  }

}
